/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.gui;

import jalview.bin.Cache;
import jalview.datamodel.Alignment;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.ColumnSelection;
import jalview.datamodel.PDBEntry;
import jalview.datamodel.SequenceI;
import jalview.ext.rbvi.chimera.JalviewChimeraBinding;
import jalview.gui.StructureViewer.ViewerType;
import jalview.io.AppletFormatAdapter;
import jalview.io.JalviewFileChooser;
import jalview.io.JalviewFileView;
import jalview.schemes.BuriedColourScheme;
import jalview.schemes.ColourSchemeI;
import jalview.schemes.HelixColourScheme;
import jalview.schemes.HydrophobicColourScheme;
import jalview.schemes.PurinePyrimidineColourScheme;
import jalview.schemes.StrandColourScheme;
import jalview.schemes.TaylorColourScheme;
import jalview.schemes.TurnColourScheme;
import jalview.schemes.ZappoColourScheme;
import jalview.structures.models.AAStructureBindingModel;
import jalview.util.BrowserLauncher;
import jalview.util.MessageManager;
import jalview.util.Platform;
import jalview.ws.dbsources.Pdb;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.Vector;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JColorChooser;
import javax.swing.JInternalFrame;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;

/**
 * GUI elements for handling an external chimera display
 * 
 * @author jprocter
 *
 */
public class ChimeraViewFrame extends StructureViewerBase
{
  private JalviewChimeraBinding jmb;

  private boolean allChainsSelected = false;

  private boolean alignAddedStructures = false;

  /*
   * state flag for PDB retrieval thread
   */
  private boolean _started = false;

  private boolean addingStructures = false;

  private IProgressIndicator progressBar = null;

  /*
   * pdb retrieval thread.
   */
  private Thread worker = null;

  /*
   * Path to Chimera session file. This is set when an open Jalview/Chimera
   * session is saved, or on restore from a Jalview project (if it holds the
   * filename of any saved Chimera sessions).
   */
  private String chimeraSessionFile = null;

  private Random random = new Random();

  /**
   * Initialise menu options.
   */
  private void initMenus()
  {
    viewerActionMenu.setText(MessageManager.getString("label.chimera"));
    viewerColour.setText(MessageManager
            .getString("label.colour_with_chimera"));
    viewerColour.setToolTipText(MessageManager
            .getString("label.let_chimera_manage_structure_colours"));
    helpItem.setText(MessageManager.getString("label.chimera_help"));
    seqColour.setSelected(jmb.isColourBySequence());
    viewerColour.setSelected(!jmb.isColourBySequence());
    if (_colourwith == null)
    {
      _colourwith = new Vector<AlignmentPanel>();
    }
    if (_alignwith == null)
    {
      _alignwith = new Vector<AlignmentPanel>();
    }

    // save As not yet implemented
    savemenu.setVisible(false);

    ViewSelectionMenu seqColourBy = new ViewSelectionMenu(
            MessageManager.getString("label.colour_by"), this, _colourwith,
            new ItemListener()
            {
              @Override
              public void itemStateChanged(ItemEvent e)
              {
                if (!seqColour.isSelected())
                {
                  seqColour.doClick();
                }
                else
                {
                  // update the Chimera display now.
                  seqColour_actionPerformed(null);
                }
              }
            });
    viewMenu.add(seqColourBy);
    viewMenu.add(fitToWindow);

    final ItemListener handler;
    JMenu alpanels = new ViewSelectionMenu(
            MessageManager.getString("label.superpose_with"), this,
            _alignwith, handler = new ItemListener()
            {
              @Override
              public void itemStateChanged(ItemEvent e)
              {
                alignStructs.setEnabled(_alignwith.size() > 0);
                alignStructs.setToolTipText(MessageManager
                        .formatMessage(
                                "label.align_structures_using_linked_alignment_views",
                                new Object[]
                                { new Integer(_alignwith.size()).toString() }));
              }
            });
    handler.itemStateChanged(null);
    viewerActionMenu.add(alpanels);
    viewerActionMenu.addMenuListener(new MenuListener()
    {

      @Override
      public void menuSelected(MenuEvent e)
      {
        handler.itemStateChanged(null);
      }

      @Override
      public void menuDeselected(MenuEvent e)
      {
        // TODO Auto-generated method stub
      }

      @Override
      public void menuCanceled(MenuEvent e)
      {
        // TODO Auto-generated method stub
      }
    });
  }

  /**
   * add a single PDB structure to a new or existing Chimera view
   * 
   * @param pdbentry
   * @param seq
   * @param chains
   * @param ap
   */
  public ChimeraViewFrame(PDBEntry pdbentry, SequenceI[] seq,
          String[] chains, final AlignmentPanel ap)
  {
    super();

    /*
     * is the pdb file already loaded?
     */
    String pdbId = pdbentry.getId();
    String alreadyMapped = ap.getStructureSelectionManager()
            .alreadyMappedToFile(pdbId);

    if (alreadyMapped != null)
    {
      int option = chooseAddSequencesToViewer(pdbId);
      if (option == JOptionPane.CANCEL_OPTION)
      {
        return;
      }
      if (option == JOptionPane.YES_OPTION)
      {
        addSequenceMappingsToStructure(seq, chains, ap, alreadyMapped);
        return;
      }
    }

    /*
     * Check if there are other Chimera views involving this alignment and give
     * user the option to add and align this molecule to one of them
     */
    List<ChimeraViewFrame> existingViews = getChimeraWindowsFor(ap);
    for (ChimeraViewFrame view : existingViews)
    {
      // TODO: highlight view somehow
      /*
       * JAL-1742 exclude view with this structure already mapped (don't offer
       * to align chain B to chain A of the same structure)
       */
      if (view.hasPdbId(pdbId))
      {
        continue;
      }
      int option = chooseAlignStructureToViewer(pdbId, view);
      if (option == JOptionPane.CANCEL_OPTION)
      {
        return;
      }
      if (option == JOptionPane.YES_OPTION)
      {
        view.useAlignmentPanelForSuperposition(ap);
        view.addStructure(pdbentry, seq, chains, true, ap.alignFrame);
        return;
      }
    }

    /*
     * If the options above are declined or do not apply, open a new viewer
     */
    openNewChimera(ap, new PDBEntry[]
    { pdbentry }, new SequenceI[][]
    { seq });
  }

  /**
   * Presents a dialog with the option to add an align a structure to an
   * existing Chimera view
   * 
   * @param pdbId
   * @param view
   * @return YES, NO or CANCEL JOptionPane code
   */
  protected int chooseAlignStructureToViewer(String pdbId,
          ChimeraViewFrame view)
  {
    int option = JOptionPane.showInternalConfirmDialog(Desktop.desktop,
            MessageManager.formatMessage("label.add_pdbentry_to_view",
                    new Object[]
                    { pdbId, view.getTitle() }), MessageManager
                    .getString("label.align_to_existing_structure_view"),
            JOptionPane.YES_NO_CANCEL_OPTION);
    return option;
  }

  /**
   * Presents a dialog with the option to add sequences to a viewer which
   * already has their structure open
   * 
   * @param pdbId
   * @return YES, NO or CANCEL JOptionPane code
   */
  protected int chooseAddSequencesToViewer(String pdbId)
  {
    int option = JOptionPane.showInternalConfirmDialog(Desktop.desktop,
            MessageManager.formatMessage(
                    "label.pdb_entry_is_already_displayed", new Object[]
                    { pdbId }), MessageManager.formatMessage(
                    "label.map_sequences_to_visible_window", new Object[]
                    { pdbId }), JOptionPane.YES_NO_CANCEL_OPTION);
    return option;
  }

  /**
   * Adds mappings for the given sequences to an already opened PDB structure,
   * and updates any viewers that have the PDB file
   * 
   * @param seq
   * @param chains
   * @param ap
   * @param pdbFilename
   */
  protected void addSequenceMappingsToStructure(SequenceI[] seq,
          String[] chains, final AlignmentPanel ap, String pdbFilename)
  {
    // TODO : Fix multiple seq to one chain issue here.
    /*
     * create the mappings
     */
    ap.getStructureSelectionManager().setMapping(seq, chains, pdbFilename,
            AppletFormatAdapter.FILE);

    /*
     * alert the FeatureRenderer to show new (PDB RESNUM) features
     */
    if (ap.getSeqPanel().seqCanvas.fr != null)
    {
      ap.getSeqPanel().seqCanvas.fr.featuresAdded();
      ap.paintAlignment(true);
    }

    /*
     * add the sequences to any other Chimera viewers for this pdb file
     */
    // JBPNOTE: this looks like a binding routine, rather than a gui routine
    for (JInternalFrame frame : Desktop.instance.getAllFrames())
    {
      if (frame instanceof ChimeraViewFrame)
      {
        ChimeraViewFrame chimeraView = ((ChimeraViewFrame) frame);
        for (int pe = 0; pe < chimeraView.jmb.getPdbCount(); pe++)
        {
          if (chimeraView.jmb.getPdbEntry(pe).getFile().equals(pdbFilename))
          {
            chimeraView.jmb.addSequence(pe, seq);
            chimeraView.addAlignmentPanel(ap);
            /*
             * add it to the set of alignments used for colouring structure by
             * sequence
             */
            chimeraView.useAlignmentPanelForColourbyseq(ap);
            chimeraView.buildActionMenu();
            ap.getStructureSelectionManager().sequenceColoursChanged(ap);
            break;
          }
        }
      }
    }
  }

  /**
   * Create a helper to manage progress bar display
   */
  protected void createProgressBar()
  {
    if (progressBar == null)
    {
      progressBar = new ProgressBar(statusPanel, statusBar);
    }
  }

  protected boolean hasPdbId(String pdbId)
  {
    return jmb.hasPdbId(pdbId);
  }

  private void openNewChimera(AlignmentPanel ap, PDBEntry[] pdbentrys,
          SequenceI[][] seqs)
  {
    createProgressBar();
    String[][] chains = extractChains(seqs);
    jmb = new JalviewChimeraBindingModel(this,
            ap.getStructureSelectionManager(), pdbentrys, seqs, chains,
            null);
    addAlignmentPanel(ap);
    useAlignmentPanelForColourbyseq(ap);
    if (pdbentrys.length > 1)
    {
      alignAddedStructures = true;
      useAlignmentPanelForSuperposition(ap);
    }
    jmb.setColourBySequence(true);
    setSize(400, 400); // probably should be a configurable/dynamic default here
    initMenus();

    addingStructures = false;
    worker = new Thread(this);
    worker.start();

    this.addInternalFrameListener(new InternalFrameAdapter()
    {
      public void internalFrameClosing(InternalFrameEvent internalFrameEvent)
      {
        closeViewer(false);
      }
    });

  }

  /**
   * Retrieve chains for sequences by inspecting their PDB refs. The hope is
   * that the first will be to the sequence's own chain. Really need a more
   * managed way of doing this.
   * 
   * @param seqs
   * @return
   */
  protected String[][] extractChains(SequenceI[][] seqs)
  {
    String[][] chains = new String[seqs.length][];
    for (int i = 0; i < seqs.length; i++)
    {
      chains[i] = new String[seqs[i].length];
      int seqno = 0;
      for (SequenceI seq : seqs[i])
      {
        String chain = null;
        if (seq.getDatasetSequence() != null)
        {
          Vector<PDBEntry> pdbrefs = seq.getDatasetSequence().getPDBId();
          if (pdbrefs != null && pdbrefs.size() > 0)
          {
            chain = pdbrefs.get(0).getChainCode();
          }
        }
        chains[i][seqno++] = chain;
      }
    }
    return chains;
  }

  /**
   * Create a new viewer from saved session state data including Chimera session
   * file
   * 
   * @param chimeraSessionFile
   * @param alignPanel
   * @param pdbArray
   * @param seqsArray
   * @param colourByChimera
   * @param colourBySequence
   * @param newViewId
   */
  public ChimeraViewFrame(String chimeraSessionFile,
          AlignmentPanel alignPanel,
          PDBEntry[] pdbArray,
          SequenceI[][] seqsArray, boolean colourByChimera,
          boolean colourBySequence, String newViewId)
  {
    super();
    setViewId(newViewId);
    this.chimeraSessionFile = chimeraSessionFile;
    openNewChimera(alignPanel, pdbArray, seqsArray);
    if (colourByChimera)
    {
      jmb.setColourBySequence(false);
      seqColour.setSelected(false);
      viewerColour.setSelected(true);
    }
    else if (colourBySequence)
    {
      jmb.setColourBySequence(true);
      seqColour.setSelected(true);
      viewerColour.setSelected(false);
    }
  }

  /**
   * create a new viewer containing several structures superimposed using the
   * given alignPanel.
   * 
   * @param pe
   * @param seqs
   * @param ap
   */
  public ChimeraViewFrame(PDBEntry[] pe, SequenceI[][] seqs,
          AlignmentPanel ap)
  {
    super();
    openNewChimera(ap, pe, seqs);
  }

  public ChimeraViewFrame(Map<PDBEntry, List<SequenceI>> toView,
          AlignmentPanel alignPanel)
  {
    super();

    /*
     * Convert the map of sequences per pdb entry into the tied arrays expected
     * by openNewChimera
     * 
     * TODO pass the Map down to openNewChimera and its callees instead
     */
    final Set<PDBEntry> pdbEntries = toView.keySet();
    PDBEntry[] pdbs = pdbEntries.toArray(new PDBEntry[pdbEntries.size()]);
    SequenceI[][] seqsForPdbs = new SequenceI[pdbEntries.size()][];
    for (int i = 0; i < pdbs.length; i++)
    {
      final List<SequenceI> seqsForPdb = toView.get(pdbs[i]);
      seqsForPdbs[i] = seqsForPdb.toArray(new SequenceI[seqsForPdb.size()]);
    }

    openNewChimera(alignPanel, pdbs, seqsForPdbs);
  }

  /**
   * add a new structure (with associated sequences and chains) to this viewer,
   * retrieving it if necessary first.
   * 
   * @param pdbentry
   * @param seq
   * @param chains
   * @param alignFrame
   * @param align
   *          if true, new structure(s) will be align using associated alignment
   */
  private void addStructure(final PDBEntry pdbentry, final SequenceI[] seq,
          final String[] chains, final boolean b,
          final IProgressIndicator alignFrame)
  {
    if (pdbentry.getFile() == null)
    {
      if (worker != null && worker.isAlive())
      {
        // a retrieval is in progress, wait around and add ourselves to the
        // queue.
        new Thread(new Runnable()
        {
          public void run()
          {
            while (worker != null && worker.isAlive() && _started)
            {
              try
              {
                Thread.sleep(100 + ((int) Math.random() * 100));

              } catch (Exception e)
              {
              }

            }
            // and call ourselves again.
            addStructure(pdbentry, seq, chains, b, alignFrame);
          }
        }).start();
        return;
      }
    }
    // otherwise, start adding the structure.
    jmb.addSequenceAndChain(new PDBEntry[]
    { pdbentry }, new SequenceI[][]
    { seq }, new String[][]
    { chains });
    addingStructures = true;
    _started = false;
    alignAddedStructures = b;
    // progressBar = alignFrame; // visual indication happens on caller frame.
    (worker = new Thread(this)).start();
    return;
  }

  private List<ChimeraViewFrame> getChimeraWindowsFor(AlignmentPanel apanel)
  {
    List<ChimeraViewFrame> result = new ArrayList<ChimeraViewFrame>();
    JInternalFrame[] frames = Desktop.instance.getAllFrames();

    for (JInternalFrame frame : frames)
    {
      if (frame instanceof ChimeraViewFrame)
      {
        if (((StructureViewerBase) frame).isLinkedWith(apanel))
        {
          result.add((ChimeraViewFrame) frame);
        }
      }
    }
    return result;
  }

  /**
   * Launch Chimera. If we have a chimera session file name, send Chimera the
   * command to open its saved session file.
   */
  void initChimera()
  {
    jmb.setFinishedInit(false);
    Desktop.addInternalFrame(this, jmb.getViewerTitle("Chimera", true),
            getBounds().width, getBounds().height);

    if (!jmb.launchChimera())
    {
      JOptionPane
              .showMessageDialog(
                      Desktop.desktop,
              MessageManager.getString("label.chimera_failed"),
              MessageManager.getString("label.error_loading_file"),
              JOptionPane.ERROR_MESSAGE);
      this.dispose();
      return;
    }

    if (this.chimeraSessionFile != null)
    {
      boolean opened = jmb.openSession(chimeraSessionFile);
      if (!opened)
      {
        System.err
                .println("An error occurred opening Chimera session file "
                        + chimeraSessionFile);
      }
    }
    jmb.setFinishedInit(true);

    jmb.startChimeraListener();
  }

  /**
   * If the list is not empty, add menu items for 'All' and each individual
   * chain to the "View | Show Chain" sub-menu. Multiple selections are allowed.
   * 
   * @param chainNames
   */
  void setChainMenuItems(List<String> chainNames)
  {
    chainMenu.removeAll();
    if (chainNames == null || chainNames.isEmpty())
    {
      return;
    }
    JMenuItem menuItem = new JMenuItem(
            MessageManager.getString("label.all"));
    menuItem.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent evt)
      {
        allChainsSelected = true;
        for (int i = 0; i < chainMenu.getItemCount(); i++)
        {
          if (chainMenu.getItem(i) instanceof JCheckBoxMenuItem)
          {
            ((JCheckBoxMenuItem) chainMenu.getItem(i)).setSelected(true);
          }
        }
        showSelectedChains();
        allChainsSelected = false;
      }
    });

    chainMenu.add(menuItem);

    for (String chainName : chainNames)
    {
      menuItem = new JCheckBoxMenuItem(chainName, true);
      menuItem.addItemListener(new ItemListener()
      {
        public void itemStateChanged(ItemEvent evt)
        {
          if (!allChainsSelected)
          {
            showSelectedChains();
          }
        }
      });

      chainMenu.add(menuItem);
    }
  }

  /**
   * Show only the selected chain(s) in the viewer
   */
  void showSelectedChains()
  {
    List<String> toshow = new ArrayList<String>();
    for (int i = 0; i < chainMenu.getItemCount(); i++)
    {
      if (chainMenu.getItem(i) instanceof JCheckBoxMenuItem)
      {
        JCheckBoxMenuItem item = (JCheckBoxMenuItem) chainMenu.getItem(i);
        if (item.isSelected())
        {
          toshow.add(item.getText());
        }
      }
    }
    jmb.showChains(toshow);
  }

  /**
   * Close down this instance of Jalview's Chimera viewer, giving the user the
   * option to close the associated Chimera window (process). They may wish to
   * keep it open until they have had an opportunity to save any work.
   * 
   * @param closeChimera
   *          if true, close any linked Chimera process; if false, prompt first
   */
  public void closeViewer(boolean closeChimera)
  {
    if (jmb.isChimeraRunning())
    {
      if (!closeChimera)
      {
        String prompt = MessageManager.formatMessage(
                "label.confirm_close_chimera", new Object[]
                { jmb.getViewerTitle("Chimera", false) });
        prompt = JvSwingUtils.wrapTooltip(true, prompt);
        int confirm = JOptionPane.showConfirmDialog(this, prompt,
                MessageManager.getString("label.close_viewer"),
                JOptionPane.YES_NO_OPTION);
        closeChimera = confirm == JOptionPane.YES_OPTION;
      }
      jmb.closeViewer(closeChimera);
    }
    setAlignmentPanel(null);
    _aps.clear();
    _alignwith.clear();
    _colourwith.clear();
    // TODO: check for memory leaks where instance isn't finalised because jmb
    // holds a reference to the window
    jmb = null;
  }

  /**
   * Open any newly added PDB structures in Chimera, having first fetched data
   * from PDB (if not already saved).
   */
  public void run()
  {
    _started = true;
    // todo - record which pdbids were successfully imported.
    StringBuilder errormsgs = new StringBuilder(128);
    StringBuilder files = new StringBuilder(128);
    List<PDBEntry> filePDB = new ArrayList<PDBEntry>();
    List<Integer> filePDBpos = new ArrayList<Integer>();
    PDBEntry thePdbEntry = null;
    try
    {
      String[] curfiles = jmb.getPdbFile(); // files currently in viewer
      // TODO: replace with reference fetching/transfer code (validate PDBentry
      // as a DBRef?)
      for (int pi = 0; pi < jmb.getPdbCount(); pi++)
      {
        String file = null;
        thePdbEntry = jmb.getPdbEntry(pi);
        if (thePdbEntry.getFile() == null)
        {
          /*
           * Retrieve PDB data, save to file, attach to PDBEntry
           */
          file = fetchPdbFile(thePdbEntry);
          if (file == null)
          {
            errormsgs.append("'" + thePdbEntry.getId() + "' ");
          }
        }
        else
        {
          /*
           * Got file already - ignore if already loaded in Chimera.
           */
          file = new File(thePdbEntry.getFile()).getAbsoluteFile()
                  .getPath();
          if (curfiles != null && curfiles.length > 0)
          {
            addingStructures = true; // already files loaded.
            for (int c = 0; c < curfiles.length; c++)
            {
              if (curfiles[c].equals(file))
              {
                file = null;
                break;
              }
            }
          }
        }
        if (file != null)
        {
          filePDB.add(thePdbEntry);
          filePDBpos.add(Integer.valueOf(pi));
          files.append(" \"" + Platform.escapeString(file) + "\"");
        }
      }
    } catch (OutOfMemoryError oomerror)
    {
      new OOMWarning("Retrieving PDB files: " + thePdbEntry.getId(),
              oomerror);
    } catch (Exception ex)
    {
      ex.printStackTrace();
      errormsgs.append("When retrieving pdbfiles for '"
              + thePdbEntry.getId() + "'");
    }
    if (errormsgs.length() > 0)
    {

      JOptionPane.showInternalMessageDialog(Desktop.desktop, MessageManager
              .formatMessage("label.pdb_entries_couldnt_be_retrieved",
                      new Object[]
                      { errormsgs.toString() }), MessageManager
              .getString("label.couldnt_load_file"),
              JOptionPane.ERROR_MESSAGE);
    }

    if (files.length() > 0)
    {
      if (!addingStructures)
      {
        try
        {
          initChimera();
        } catch (Exception ex)
        {
          Cache.log.error("Couldn't open Chimera viewer!", ex);
        }
      }
      int num = -1;
      for (PDBEntry pe : filePDB)
      {
        num++;
        if (pe.getFile() != null)
        {
          try
          {
            int pos = filePDBpos.get(num).intValue();
            long startTime = startProgressBar("Chimera "
                    + MessageManager.getString("status.opening_file"));
            jmb.openFile(pe);
            jmb.addSequence(pos, jmb.getSequence()[pos]);
            File fl = new File(pe.getFile());
            String protocol = AppletFormatAdapter.URL;
            try
            {
              if (fl.exists())
              {
                protocol = AppletFormatAdapter.FILE;
              }
            } catch (Throwable e)
            {
            } finally
            {
              stopProgressBar("", startTime);
            }
            // Explicitly map to the filename used by Chimera ;
            jmb.getSsm().setMapping(jmb.getSequence()[pos],
                    jmb.getChains()[pos],
                    pe.getFile(),
                    protocol);
          } catch (OutOfMemoryError oomerror)
          {
            new OOMWarning(
                    "When trying to open and map structures from Chimera!",
                    oomerror);
          } catch (Exception ex)
          {
            Cache.log.error("Couldn't open " + pe.getFile()
                    + " in Chimera viewer!", ex);
          } finally
          {
            Cache.log.debug("File locations are " + files);
          }
        }
      }
      jmb.setFinishedInit(true);
      jmb.setLoadingFromArchive(false);

      // refresh the sequence colours for the new structure(s)
      for (AlignmentPanel ap : _colourwith)
      {
        jmb.updateColours(ap);
      }
      // do superposition if asked to
      if (alignAddedStructures)
      {
        new Thread(new Runnable()
        {
          public void run()
          {
            alignStructs_withAllAlignPanels();
          }
        }).start();
        alignAddedStructures = false;
      }
      addingStructures = false;
    }
    _started = false;
    worker = null;
  }

  /**
   * Fetch PDB data and save to a local file. Returns the full path to the file,
   * or null if fetch fails.
   * 
   * @param processingEntry
   * @return
   * @throws Exception
   */
  private String fetchPdbFile(PDBEntry processingEntry) throws Exception
  {
    String filePath = null;
    Pdb pdbclient = new Pdb();
    AlignmentI pdbseq = null;
    String pdbid = processingEntry.getId();
    long handle = System.currentTimeMillis()
            + Thread.currentThread().hashCode();

    /*
     * Write 'fetching PDB' progress on AlignFrame as we are not yet visible
     */
    String msg = MessageManager.formatMessage("status.fetching_pdb",
            new Object[]
            { pdbid });
    getAlignmentPanel().alignFrame.setProgressBar(msg, handle);
    // long hdl = startProgressBar(MessageManager.formatMessage(
    // "status.fetching_pdb", new Object[]
    // { pdbid }));
    try
    {
      pdbseq = pdbclient.getSequenceRecords(pdbid);
    } catch (OutOfMemoryError oomerror)
    {
      new OOMWarning("Retrieving PDB id " + pdbid, oomerror);
    } finally
    {
      msg = pdbid + " "
              + MessageManager.getString("label.state_completed");
      getAlignmentPanel().alignFrame.setProgressBar(msg, handle);
      // stopProgressBar(msg, hdl);
    }
    /*
     * If PDB data were saved and are not invalid (empty alignment), return the
     * file path.
     */
    if (pdbseq != null && pdbseq.getHeight() > 0)
    {
      // just use the file name from the first sequence's first PDBEntry
      filePath = new File(pdbseq.getSequenceAt(0).getPDBId()
              .elementAt(0).getFile()).getAbsolutePath();
      processingEntry.setFile(filePath);
    }
    return filePath;
  }

  /**
   * Convenience method to update the progress bar if there is one. Be sure to
   * call stopProgressBar with the returned handle to remove the message.
   * 
   * @param msg
   * @param handle
   */
  public long startProgressBar(String msg)
  {
    // TODO would rather have startProgress/stopProgress as the
    // IProgressIndicator interface
    long tm = random.nextLong();
    if (progressBar != null)
    {
      progressBar.setProgressBar(msg, tm);
    }
    return tm;
  }

  /**
   * End the progress bar with the specified handle, leaving a message (if not
   * null) on the status bar
   * 
   * @param msg
   * @param handle
   */
  public void stopProgressBar(String msg, long handle)
  {
    if (progressBar != null)
    {
      progressBar.setProgressBar(msg, handle);
    }
  }

  @Override
  public void pdbFile_actionPerformed(ActionEvent actionEvent)
  {
    JalviewFileChooser chooser = new JalviewFileChooser(
            Cache.getProperty("LAST_DIRECTORY"));

    chooser.setFileView(new JalviewFileView());
    chooser.setDialogTitle(MessageManager.getString("label.save_pdb_file"));
    chooser.setToolTipText(MessageManager.getString("action.save"));

    int value = chooser.showSaveDialog(this);

    if (value == JalviewFileChooser.APPROVE_OPTION)
    {
      BufferedReader in = null;
      try
      {
        // TODO: cope with multiple PDB files in view
        in = new BufferedReader(new FileReader(jmb.getPdbFile()[0]));
        File outFile = chooser.getSelectedFile();

        PrintWriter out = new PrintWriter(new FileOutputStream(outFile));
        String data;
        while ((data = in.readLine()) != null)
        {
          if (!(data.indexOf("<PRE>") > -1 || data.indexOf("</PRE>") > -1))
          {
            out.println(data);
          }
        }
        out.close();
      } catch (Exception ex)
      {
        ex.printStackTrace();
      } finally
      {
        if (in != null)
        {
          try
          {
            in.close();
          } catch (IOException e)
          {
            e.printStackTrace();
          }
        }
      }
    }
  }

  @Override
  public void viewMapping_actionPerformed(ActionEvent actionEvent)
  {
    CutAndPasteTransfer cap = new CutAndPasteTransfer();
    try
    {
      cap.appendText(jmb.printMappings());
    } catch (OutOfMemoryError e)
    {
      new OOMWarning(
              "composing sequence-structure alignments for display in text box.",
              e);
      cap.dispose();
      return;
    }
    Desktop.addInternalFrame(cap,
            MessageManager.getString("label.pdb_sequence_mapping"), 550,
            600);
  }

  @Override
  public void eps_actionPerformed(ActionEvent e)
  {
    throw new Error(
            MessageManager
                    .getString("error.eps_generation_not_implemented"));
  }

  @Override
  public void png_actionPerformed(ActionEvent e)
  {
    throw new Error(
            MessageManager
                    .getString("error.png_generation_not_implemented"));
  }

  @Override
  public void viewerColour_actionPerformed(ActionEvent actionEvent)
  {
    if (viewerColour.isSelected())
    {
      // disable automatic sequence colouring.
      jmb.setColourBySequence(false);
    }
  }

  @Override
  public void seqColour_actionPerformed(ActionEvent actionEvent)
  {
    jmb.setColourBySequence(seqColour.isSelected());
    if (_colourwith == null)
    {
      _colourwith = new Vector<AlignmentPanel>();
    }
    if (jmb.isColourBySequence())
    {
      if (!jmb.isLoadingFromArchive())
      {
        if (_colourwith.size() == 0 && getAlignmentPanel() != null)
        {
          // Make the currently displayed alignment panel the associated view
          _colourwith.add(getAlignmentPanel().alignFrame.alignPanel);
        }
      }
      // Set the colour using the current view for the associated alignframe
      for (AlignmentPanel ap : _colourwith)
      {
        jmb.colourBySequence(ap.av.isShowSequenceFeatures(), ap);
      }
    }
  }

  @Override
  public void chainColour_actionPerformed(ActionEvent actionEvent)
  {
    chainColour.setSelected(true);
    jmb.colourByChain();
  }

  @Override
  public void chargeColour_actionPerformed(ActionEvent actionEvent)
  {
    chargeColour.setSelected(true);
    jmb.colourByCharge();
  }

  @Override
  public void zappoColour_actionPerformed(ActionEvent actionEvent)
  {
    zappoColour.setSelected(true);
    jmb.setJalviewColourScheme(new ZappoColourScheme());
  }

  @Override
  public void taylorColour_actionPerformed(ActionEvent actionEvent)
  {
    taylorColour.setSelected(true);
    jmb.setJalviewColourScheme(new TaylorColourScheme());
  }

  @Override
  public void hydroColour_actionPerformed(ActionEvent actionEvent)
  {
    hydroColour.setSelected(true);
    jmb.setJalviewColourScheme(new HydrophobicColourScheme());
  }

  @Override
  public void helixColour_actionPerformed(ActionEvent actionEvent)
  {
    helixColour.setSelected(true);
    jmb.setJalviewColourScheme(new HelixColourScheme());
  }

  @Override
  public void strandColour_actionPerformed(ActionEvent actionEvent)
  {
    strandColour.setSelected(true);
    jmb.setJalviewColourScheme(new StrandColourScheme());
  }

  @Override
  public void turnColour_actionPerformed(ActionEvent actionEvent)
  {
    turnColour.setSelected(true);
    jmb.setJalviewColourScheme(new TurnColourScheme());
  }

  @Override
  public void buriedColour_actionPerformed(ActionEvent actionEvent)
  {
    buriedColour.setSelected(true);
    jmb.setJalviewColourScheme(new BuriedColourScheme());
  }

  @Override
  public void purinePyrimidineColour_actionPerformed(ActionEvent actionEvent)
  {
    setJalviewColourScheme(new PurinePyrimidineColourScheme());
  }

  @Override
  public void userColour_actionPerformed(ActionEvent actionEvent)
  {
    userColour.setSelected(true);
    new UserDefinedColours(this, null);
  }

  @Override
  public void backGround_actionPerformed(ActionEvent actionEvent)
  {
    java.awt.Color col = JColorChooser
            .showDialog(this, MessageManager
                    .getString("label.select_backgroud_colour"), null);
    if (col != null)
    {
      jmb.setBackgroundColour(col);
    }
  }

  @Override
  public void showHelp_actionPerformed(ActionEvent actionEvent)
  {
    try
    {
      BrowserLauncher
              .openURL("https://www.cgl.ucsf.edu/chimera/docs/UsersGuide");
    } catch (Exception ex)
    {
    }
  }

  public void updateTitleAndMenus()
  {
    if (jmb.fileLoadingError != null && jmb.fileLoadingError.length() > 0)
    {
      repaint();
      return;
    }
    setChainMenuItems(jmb.getChainNames());

    this.setTitle(jmb.getViewerTitle("Chimera", true));
    if (jmb.getPdbFile().length > 1 && jmb.getSequence().length > 1)
    {
      viewerActionMenu.setVisible(true);
    }
    if (!jmb.isLoadingFromArchive())
    {
      seqColour_actionPerformed(null);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see
   * jalview.jbgui.GStructureViewer#alignStructs_actionPerformed(java.awt.event
   * .ActionEvent)
   */
  @Override
  protected void alignStructs_actionPerformed(ActionEvent actionEvent)
  {
    alignStructs_withAllAlignPanels();
  }

  private void alignStructs_withAllAlignPanels()
  {
    if (getAlignmentPanel() == null)
    {
      return;
    }

    if (_alignwith.size() == 0)
    {
      _alignwith.add(getAlignmentPanel());
    }

    try
    {
      AlignmentI[] als = new Alignment[_alignwith.size()];
      ColumnSelection[] alc = new ColumnSelection[_alignwith.size()];
      int[] alm = new int[_alignwith.size()];
      int a = 0;

      for (AlignmentPanel ap : _alignwith)
      {
        als[a] = ap.av.getAlignment();
        alm[a] = -1;
        alc[a++] = ap.av.getColumnSelection();
      }
      jmb.superposeStructures(als, alm, alc);
    } catch (Exception e)
    {
      StringBuffer sp = new StringBuffer();
      for (AlignmentPanel ap : _alignwith)
      {
        sp.append("'" + ap.alignFrame.getTitle() + "' ");
      }
      Cache.log.info("Couldn't align structures with the " + sp.toString()
              + "associated alignment panels.", e);
    }
  }

  public void setJalviewColourScheme(ColourSchemeI ucs)
  {
    jmb.setJalviewColourScheme(ucs);

  }

  /**
   * 
   * @param alignment
   * @return first alignment panel displaying given alignment, or the default
   *         alignment panel
   */
  public AlignmentPanel getAlignmentPanelFor(AlignmentI alignment)
  {
    for (AlignmentPanel ap : getAllAlignmentPanels())
    {
      if (ap.av.getAlignment() == alignment)
      {
        return ap;
      }
    }
    return getAlignmentPanel();
  }

  @Override
  public AAStructureBindingModel getBinding()
  {
    return jmb;
  }

  /**
   * Ask Chimera to save its session to the designated file path, or to a
   * temporary file if the path is null. Returns the file path if successful,
   * else null.
   * 
   * @param filepath
   * @see getStateInfo
   */
  protected String saveSession(String filepath)
  {
    String pathUsed = filepath;
    try
    {
      if (pathUsed == null)
      {
        File tempFile = File.createTempFile("chimera", ".py");
        tempFile.deleteOnExit();
        pathUsed = tempFile.getPath();
      }
      boolean result = jmb.saveSession(pathUsed);
      if (result)
      {
        this.chimeraSessionFile = pathUsed;
        return pathUsed;
      }
    } catch (IOException e)
    {
    }
    return null;
  }

  /**
   * Returns a string representing the state of the Chimera session. This is
   * done by requesting Chimera to save its session to a temporary file, then
   * reading the file contents. Returns an empty string on any error.
   */
  @Override
  public String getStateInfo()
  {
    String sessionFile = saveSession(null);
    if (sessionFile == null)
    {
      return "";
    }
    InputStream is = null;
    try
    {
      File f = new File(sessionFile);
      byte[] bytes = new byte[(int) f.length()];
      is = new FileInputStream(sessionFile);
      is.read(bytes);
      return new String(bytes);
    } catch (IOException e)
    {
      return "";
    } finally
    {
      if (is != null)
      {
        try
        {
          is.close();
        } catch (IOException e)
        {
          // ignore
        }
      }
    }
  }

  @Override
  protected void fitToWindow_actionPerformed()
  {
    jmb.focusView();
  }

  @Override
  public ViewerType getViewerType()
  {
    return ViewerType.CHIMERA;
  }
}
